"""A recipe example for preprocessing Finnish parliament transcripts to a kaldi text file."""
import re
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Match
from typing import Tuple

import aalto_asr_preprocessor.fi.numbers.expansion as num_fi

SUFFIX_MAPPING: Dict[str, List[str]] = {
    "NOM": ["han", "hän", "kaan", "kään", "ko", "kö", "kin", "pa", "pä"],
    "PAR": ["a", "ä", "aa", "ää", "ta", "taa", "tta", "tä"],
    "GEN": ["n"],
    "INE": ["ssa", "ssä"],
    "ELA": ["sta", "stä"],
    "ILL": ["aan", "een", "iin", "teen", "ään"],
    "ADE": ["lla", "llä"],
    "ABL": ["lta", "ltä"],
    "ALL": ["lle"],
    "ESS": ["na", "nä"],
    "TRA": ["ksi"],
    "JNOM": ["nen", "s"],
    "JPAR": [
        "ttä"
    ],  # overlaps with some #PAR cases, due to rarity overlaps are not handled
    "JGEN": ["nnen"],
    "JINE": ["nnessa", "nnessä"],
    "JELA": ["nnesta", "nnestä"],
    "JILL": ["nteen"],
    "JADE": ["nnella", "nnellä"],
    "JABL": ["nnelta", "nneltä"],
    "JALL": ["nnelle"],
    "JESS": ["ntena", "ntenä"],
    "JTRA": ["nneksi"],
}

WORD_INFLECTIONS: List[Tuple[str, str]] = [
    # 100 foobariksi -> sadaksi
    (r"[a-zåäö]{2,}ksi$", "TRA"),
    # 5 foobarilta -> viideltä
    (r"ll[aä]$", "ADE"),
    # 5 foobarilla -> viidellä
    (r"lt[aä]$", "ABL"),
    # 5 foobarille -> viidelle
    (r"lle$", "ALL"),
    # 5 foobarina -> viitenä
    (r"n[aä]$", "ESS"),
    # 6 foobarissa -> kuudessa
    (r"ss[aä]$", "INE"),
    # 6 foobarista -> kuudesta
    (r"st[aä]$", "ELA"),
    # 10 foobariin -> kymmeneen (require at least five chars because otherwise genitive cases of
    # short words like maa [maan], muu [muun] etc. get wrong case here)
    (r"[a-zåäö]{2,}(?:aa|ee|ii|oo|uu|yy|ää|öö)n$", "ILL"),
    # 3 foobarin -> kolmen (require at least three chars so 'on' is not mistakenly caught here)
    # Needs to be last, otherwise catches words in illative case too.
    (r"[a-zåäö]{2,}n$", "GEN"),
]

ELATIVE_WORDS = (
    Path("src/aalto_asr_preprocessor/fi/words_elative.txt").read_text().split()
)

SECTION_CHAR_MAPPING: Dict[str, Tuple[str, str]] = {
    "": ("pykälä", "JNOM"),
    "hän": ("pykälä", "JNOM"),
    "kin": ("pykälä", "JNOM"),
    "ää": ("pykäl", "JPAR"),
    "n": ("pykälä", "JGEN"),
    "ssä": ("pykälä", "JINE"),
    "ssään": ("pykälä", "JINE"),
    "stä": ("pykälä", "JELA"),
    "ään": ("pykäl", "JILL"),
    "llä": ("pykälä", "JADE"),
    "lle": ("pykälä", "JALL"),
    "nä": ("pykälä", "JESS"),
    "ksi": ("pykälä", "JTRA"),
    "t": ("pykälä", "NOM"),
    "iä": ("pykäl", "NOM"),
    "ien": ("pykäl", "NOM"),
    "iin": ("pykäl", "NOM"),
    "issä": ("pykäl", "NOM"),
    "s": ("pykälä", "NOM"),  # In swedish text, include here to prevent KeyError
}

ROMAN_TO_NUMBER: Dict[str, str] = {
    "I": "1",
    "II": "2",
    "III": "3",
    "IV": "4",
    "V": "5",
    "VI": "6",
    "VII": "7",
    "VIII": "8",
    "IX": "9",
    "X": "10",
}


def map_suffix_to_inflection(match: Match[str]) -> str:
    """Expand inflected numbers captured with regexp.

    Can handle single numbers, decimal numbers and ranges. Possible clitics are handled too.
        3:sta -> kolmesta
        4,4:ään -> neljään pilkku neljään
        1 000-1 500:ssa -> tuhannessa viiva tuhannessaviidessäsadassa
        20:nkin -> kahdenkymmenenkin

    Args:
        match (Match): number(s) and inflection captured using regexp

    Returns:
        str: number(s) inflected in word format
    """
    numbers = (
        match.group(2) if match.group(1) is None else match.group(1) + match.group(2)
    )
    suffix, clitic = match.group(3), ""
    form = "NOM"
    if tmp := [
        (suffix.split(cl)[0], cl) for cl in SUFFIX_MAPPING["NOM"] if cl in suffix
    ]:
        suffix, clitic = tmp[0]
    for f, suffixes in SUFFIX_MAPPING.items():
        if suffix in suffixes:
            form = f
    translation = str.maketrans(
        {"+": " plus ", ",": " pilkku ", ".": " piste ", "/": " per "}
    )
    numbers = numbers.translate(translation)
    numbers = re.sub(r"[-–—]", r" viiva ", numbers)
    numbers = re.sub(r"(\d+)\b", r"\1" + form, numbers)
    numbers = re.sub(r"(\d+)([A-Z]+)?", num_fi.expand, numbers)
    return f"{numbers}{clitic}"


def number_word_pair_inflection(match: Match[str]) -> Any:
    """Inflect number with the same case as following word.

    Elative overlaps with partitive which is solved using a separate lookup table collected from
    Finnish parliament text data. For other inflections, it is assumed that the suffix of the word
    informs the correct inflection for the number. Not 100% accurate/valid assumption but it more
    probable than leaving all numbers in nominal form.

    Args:
        match (Match): number(s) and word captured using regexp

    Returns:
        Any: captured regex with numbers inflected in word format
    """
    result, word = match.group(0), match.group(3)
    for regex, form in WORD_INFLECTIONS:
        if re.search(regex, word):
            if form == "ELA" and word not in ELATIVE_WORDS:
                form = "NOM"
            result = re.sub(r"(\d+\.?)", r"\1#" + form, result)
            break
    return re.sub(r"(\d+)\.([A-Z]+\b)", r"\1#J\2", result)


def change_to_ordinal(match: Match[str]) -> str:
    """Change number before specific captured words to ordinal form.

    Usually ordinals are marked like so: 4. päivänä -> neljäntenä päivänä. But periods are left
    out from numbers before words such as 'momentti' or 'kohta' even if the number is read as
    ordinal. However, if these words are preceded by a list of numbers then they are here expanded
    in nominal case because MPs are more likely to read them in nominal form.

    Args:
        match (Match): number(s) and word captured using regexp

    Returns:
        str: number(s) inflected in word format together with the following word
    """
    numbers = (
        match.group(2) if match.group(1) is None else match.group(1) + match.group(2)
    )
    form = match.group(3) if match.group(3) is not None else "NOM"
    word = match.group(4)
    numbers = re.sub(r"[-–—]", r" viiva ", numbers)
    if match.group(1):
        numbers = re.sub(r"(\d+),?", r"\1NOM", numbers)
    else:
        numbers = re.sub(r"(\d+)", r"\1J" + form, numbers)
    numbers = re.sub(r"(\d+)([A-Z]+)?", num_fi.expand, numbers)
    return f"{numbers} {word}"


def expand_section_sign(match: Match[str]) -> str:
    """Handle the inflection of § sign and the numbers immediately preceding it.

    Determines the inflection from the suffix associated with §. If there are multiple numbers, then
    all numbers are expanded in nominal form because MPs are more likely read it that way.

    Args:
        match (Match): number(s) and the inflection of the section sign

    Returns:
        str: number(s) and the section sign expanded in the correct inflected form
    """
    numbers = (
        match.group(2) if match.group(1) is None else match.group(1) + match.group(2)
    )
    suffix, clitic = match.group(3), ""
    if tmp := [
        (suffix.split(cl)[0], cl) for cl in SUFFIX_MAPPING["NOM"] if cl in suffix
    ]:
        suffix, clitic = tmp[0]
    section_word, form = SECTION_CHAR_MAPPING[suffix]
    numbers = re.sub(r"[-–—]", r" viiva ", numbers)
    numbers = re.sub(r"(\d+)([a-z])", r"\1 \2", numbers)
    numbers = re.sub(r"(\d+),?", r"\1" + form, numbers)
    numbers = re.sub(r"(\d+)([A-Z]+)?", num_fi.expand, numbers)
    if form == "NOM":
        return f"{section_word}{suffix} {numbers}"
    return f"{numbers} {section_word}{suffix}{clitic}"


def school_classes(match: Match[str]) -> Any:
    """Handle the specific reading of 1st and 2nd year school classes.

    Args:
        match (Match): captured school class(es)

    Returns:
        Any: school classes expanded in correct inflected form
    """
    result = " ".join([i or "" for i in match.groups()]).strip()
    translation = str.maketrans({"1": "ykkös", "2": "kakkos", "—": "viiva"})
    result = result.translate(translation)
    result = re.sub(r"(\d+)\b", r"\1JNOM", result)
    return re.sub(r"(\d+)([A-Z]+)?", num_fi.expand, result)


def roman_numerals(match: Match[str]) -> Any:
    """Map roman numerals to arabic numerals. Handles currently only 1 to 10.

    Args:
        match (Match): captured roman numeral

    Returns:
        Any: roman numeral as its corresponding arabic numeral
    """
    result, roman = match.group(0), match.group(2)
    number = ROMAN_TO_NUMBER[roman]
    result = result.replace(roman, number)
    return re.sub(r"(\d+)-", r"\1. ", result)


def years(match: Match[str]) -> Any:
    """Handle years when they have only two digits.

    Args:
        match (Match): captured years and words connecting them

    Returns:
        Any: year inflected in the nominal form
    """
    result = match.group(0)
    result = re.sub(r"[-–—]", r" viiva ", result)
    result = re.sub(r"(\d+)", r"\1NOM", result)
    return re.sub(r"(\d+)([A-Z]+)", num_fi.expand, result)


def lowercase(match: Match[str]) -> Any:
    """Convert match to lowercase.

    Args:
        match (Match): captured string

    Returns:
        Any: capture in lowercase
    """
    return match.group(0).lower()


UNACCEPTED_CHARS = r"[^a-zåäö \n]"

REGEXPS = [
    (r"\n", r" "),
    # Initials
    (r"\b([A-ZÄÖÅ])\.", r"\1 "),
    # Välihuudot pois = Remove shouts and interruptions
    (r"\s*(?:\(|\[).*?(?:\)|\])\s*", r" "),
    # Typos that have been encountered: Unmatched bracket in the beginning or unmatched bracket in
    # an interruption at the end
    (r"^(?:\(|\[)", r""),
    (r"(?:\(|\[)[^)\]]*?$", r" "),
    # (r"\s*\[.*?\]\s*", r" "),
    # Urlit pois = Remove URLs
    (r"http:\S*", r""),
    (r"www\.[a-zA-Z]\S*", r""),
    (r"\S*\.html?", r""),
    # Peräkkäiset pisteet yhdeksi = Substitute repeated periods with a single period
    (r"\.+", r"."),
    # Vuosiluvut = Years
    (
        r"(?<!ensi )(?<!viime )([Vv]uo[nds]\w{1,5} )(\d{2,4}(?:—| ja | tai ))?(\d{2})(?!.?\d)",
        years,
    ),
    (r"(\d{4})[–—]", r"\1 viiva "),
    (r"([^\/])(\d{4})\b", r"\1\2#NOM"),
    # Välit pois numeroiden välistä = Remove spaces between digits
    (r"(\d)\s+(\d)", r"\1\2"),
    # Roomalaiset numerot 1-10 = Roman numerals between 1 and 10
    (r"( |-)(X|IX|IV|V|V?I{1,3})( |-|:|\.)", roman_numerals),
    # Jotain erikoissanontoja = Some abbreviations
    (r"\bjr.", r"junior "),
    (r"\binc.", r"inc "),
    (r"(\d+)\s*m/s", r"\1 metriä sekunnissa"),
    (r"(\d+)\s*r/min", r"\1 kierrosta minuutissa"),
    (r"(\d+)\s+kpm\b", r"\1 kilopondimetriä"),
    (r"(\d+)\s+kW/l\b", r"\1 kilowattia litraa kohti"),
    (r"(\d+)\s+kW\b", r"\1 kilowattia"),
    (r"(\d+)\s+hv/l\b", r"\1 hevosvoimaa litraa kohti"),
    (r"1½", r" yksi ja puoli "),
    (r"½", r" puoli "),
    (r"m²:n", r"neliömetrin"),
    (r"\/\s*m²", r" per neliömetri"),
    (r"\S*[²³]\S*", r""),
    (r"±", r"plus miinus "),
    (r"(\d+)\s*°C", r"\1 Celcius astetta"),
    (r"(\d+)\s*°", r"\1 astetta"),
    (
        r"(?:klo|[kK]ello)\s+(\d+)\.(\d+)—(\d+)\.(\d+)",
        r"kello \1#NOM \2#NOM viiva \3#NOM \4#NOM",
    ),
    (r"(?:klo|[kK]ello)\s+(on\s+)?(\d+)\.(\d+)?", r"kello \1 \2#NOM \3"),
    (r"£", r"puntaa"),
    (r"\$", r"dollaria"),
    (r"\%", r"prosenttia"),
    (r"\besim\.", r"esimerkiksi "),
    (r"\bpat\. no\.?", r"patentti numero "),
    (r"\bo\.s\.", r"omaa sukua "),
    (r"\s+n\.", r" noin"),
    (r"\bts\.", r"tai siis "),
    (r"\bev\. ?lut\.", r"evankelisluterilainen "),
    (r"\bpros\.", r"prosenttia"),
    (r"\bk\.?o\.", r"kyseessä oleva "),
    (r"\bjne.", r"ja niin edelleen "),
    (r"\bem\.", r"edellä mainittu "),
    (r"\bent\.", r"entinen "),
    (r"\bop\.\s*(\d+)", r"opus \1 "),
    (r"\bkv\.", r"kansainvälinen "),
    (r"\s+ns\.", r" niin sanottu "),
    (r"\s+nk\.", r" niin kutsuttu "),
    (r"\s+vt\.", r" virkaatekevä "),
    (r"\s+p\.\s+(\d+)", r" puhelin \1"),
    (r"\bpuh\.\s+(\d+)", r"puhelin \1"),
    (r"\s+v\.\s+(\d+)", r" vuonna \1"),
    (r"\s+(s|synt)\.\s*(\d{4})", r" syntynyt \1"),
    (r"\bym\.", r"ynnä muuta "),
    (r"\byms\.", r"ynnä muuta sellaista "),
    (r"\boy\.", r"osakeyhtiö "),
    (r"\s+ry\.", r" rekisteröity yhdistys "),
    (r"\s+mm\.", r" muun muassa "),
    (r"\bmrd\. mk\.", r"miljardia markkaa "),
    (r"\bmrd\.", r"miljardia "),
    (r"\bmilj\.", r"miljoonaa "),
    (r"\bvas\.", r"vasemmistoliitto "),
    (r"\bvihr\.", r"vihreä liitto "),
    (r"\bV65\b", r"vee kuusi viisi "),
    (r"\be. ?Kr.", r"ennen kristusta "),
    (r"\beKr\.?", r"ennen kristusta "),
    (r"\bks\. ?ed.", r"kansanedustaja "),
    (r"\bpj\. ?ed.", r"puheenjohtaja "),
    (r"\b[Tt]oim\. ?joht.", r"toimitusjohtaja "),
    (r"\b[Oo]pett\. ?", r"opettaja "),
    (r"\b[Pp]uh\. ?joht.", r"puheenjohtaja "),
    (r"\b[Kk]auppat\.", r"kauppatieteiden "),
    (r"\b[Kk]asvatustiet\.", r"kasvatustieteiden "),
    (r"\b[Vv]altiotiet\.", r"valtiotieteiden "),
    (r"\b[Ll]ääket\.", r"lääketieteen "),
    (r"\b[Mm]aist\.", r"maisteri "),
    (r"\b[Ll]is\.", r"lisensiaatti "),
    (r"(\d+)\s+hv\b", r"\1 hevosvoimaa"),
    (r"(\d+)\s+g\s+(?!§)", r"\1 grammaa"),
    (r"(\d+)\s+mg\b", r"\1 milligrammaa"),
    (r"(\d+)\s+dl\b", r"\1 desilitraa"),
    (r"(\d+)\s+l\b", r"\1 litraa"),
    (r"(\d+)\s+bar\b", r"\1 baaria"),
    (r"(\d+)\s+kg\b", r"\1 kilogrammaa"),
    (r"(\d+)\s+km\b", r"\1 kilometriä"),
    (r"(\d+)\s+kk\b", r"\1 kuukautta"),
    (r"(\d+)\s+mm\b", r"\1 millimetriä"),
    (r"(\d+)\s+cm\b", r"\1 senttimetriä"),
    (r"(\d+)\s+min\b", r"\1 minuuttia"),
    (r"(\d+)\s+m\b", r"\1 metriä"),
    (r"(\d+)\s+h\b", r"\1 tuntia"),
    (r"(\d+)\s+mk\b", r"\1 markkaa"),
    (r"(\d+)\s*v\b", r"\1 vuotta"),
    (r"\.000(?=\D)", r"000"),
    (r"-(\d+)\s", r"miinus \1"),
    (r" (\d)\.0", r" \1 piste nolla"),
    # Eduskunta-aineiston erikoistapauksia = Some parliament data special cases
    (r"AM/121", r"am yksi kaksi yksi"),
    (r"112-", r"yksi yksi kaksi "),
    (r"Sr2:ia", r"sr kakkosia"),
    (r"Sr3:ia", r"sr kolmosia"),
    # Separate letter-digit codes which begin with a capital letter (H5N8 -> H 5N 8, K18 -> K 18)
    # [the '5N' is separated by a later regex]
    (r"([a-zA-Z]+)(\d+)", r"\1 \2"),
    # Kaikki pykälämerkit ja niihen liittyvät numerot = Catch section symbol and all
    # associated numbers
    (
        r"((?:\d+(?: [a-z])?, )*\d+(?: [a-z])?(?:—|-| ja ))?(\d+(?: *[a-z])?) *§:?([a-zåäö]*)",
        expand_section_sign,
    ),
    # Ruotsinkielisessä tekstissä pykälämerkki tulee ennen numeroa. Käsitellään toistaiseksi näin
    # koska pelkällä poistolla virheet edellisessä regexpissä jäävät huomaamatta = This format
    # appears in Swedish texts, handle like this for now so errors with above regexp aren't missed
    (r"§ (\d+)", r"\1"),
    # Taivutetut numerot = Handle numbers with inflections attached
    (
        r"(\d+[/+,.–—-])?(\d+):([a-zåäö]+)\b",
        map_suffix_to_inflection,
    ),
    # Ykkös- ja kakkosluokkalaisten tapaus = Exception in pronouncing school classes
    (
        r"(?:(\d+)[. -]*(—|ja )?)?(\d+)[. ]*-(luokkal)",
        school_classes,
    ),
    # Numerovälit = Number ranges
    (r"((?:\d+[. ]*)+)[–—-]+((?:\d+[. ]*)+)", r"\1 viiva \2"),
    # Eduskunnan dokumenttinumerot = Document numbers in parliament (the second number = year
    # and the committee abbreviation in the beginning are usually not pronounced)
    (r"[A-Z][a-zA-Z]{1,2}[A-Z]{0,2} (\d+)\/\d+(?: vp)?", r" \1#NOM"),
    # Budjetin kohta
    (r"(moment[a-z]+) (\d{2})\.(\d{2})\.(\d{2})", r"\1 \2#NOM \3#NOM \4#NOM"),
    # Päivämäärät = Dates
    (r"(\d+)\s*\.\s*(\d+)\s*\.", r" \1#JNOM \2#JPAR "),
    # Numerot joissa välimerkkejä = Numbers with punctuation
    (r"(\d+)\.(\d+)", r"\1 piste \2"),
    (r"(\d+),(\d+)", r"\1 pilkku \2"),
    (r"(\d+)\/(\d+)", r"\1 kautta \2"),
    (r"\+", r" plus "),
    (r"(\d+):(\d+)", r"\1#GEN suhde \2#ILL"),
    # Numeroitu luettelo = Numbered list item
    (r"(\d+)\)", r"\1#NOM"),
    # Separate codes which start with a digit and are combined with combined capital letters.
    # May also contain inflections (3G:tä -> 3 G tä, 3D:ssä -> 3 D ssä)
    (r"(\d)([A-Za-zåäö]+):?([a-zåäö]+)?", r"\1#NOM \2 \3"),
    # Separate when number and word are accidentally missing a space between
    (r"(\d+)([a-z]+)", r"\1 \2"),
    # Nollalla alkavat numerot erikseen = Separate numbers that begin with a zero
    (r"(?<![0-9])(0\d*)(#[A-Z#]+)?", r"\1#ERI"),
    # Expand numbers to correct case
    (
        r"(\d+\.? (?:viiva|pilkku|piste) )?(\d+\.?)\s+([A-Za-zåäö-]+)",
        number_word_pair_inflection,
    ),
    # Muuta järjestyslukujen sijamuodot = Change to ordinal if number is followed by a period
    (r"(\d+)\.\#([A-Z]+\b)", r"\1#J\2"),
    # 12. [a-z] -> kahdestoista
    (r"(\d+)\.\s+([a-zäöå])", r"\1#JNOM \2"),
    # Sanaa momentti ja kohta edeltävät luvut ovat järjestysnumeroita vaikka niissä ei ole pistettä
    (
        r"((?:\d+, )*\d+(?:—|-| ja ))?(\d+)#?([A-Z]{3})?\s+(moment|koh[dt]|artikl)",
        change_to_ordinal,
    ),
    # Lavenna jäljellä olevat numerot = Expand remaining numbers
    (r"(\d+)\#?([A-Z]+)?", num_fi.expand),
    # Väliviivat sanarajoiksi = Use hyphens as word boundaries
    (r"[–—‑-]+", r" "),
    # Ilmiselviä välimerkkejä lauserajoiksi = Turn obvious punctuation to sentence boundaries
    (r"[.:?!]\s+", r"\n"),
    # Muut välimerkit = Other punctuation
    (r"[!\?;…\/]", r"\n"),
    (r"[>¤¶†ªðÐº¨¦¾Þ\\©®þ\«,­¸:_\»<=&\*()\]¿¡#@~\"'`´‘’“”]", r""),
    (r"\s?—\s?", r" "),
    # Loput pisteet pois = Remove rest of periods
    (r"\.", r" "),
    # Strip any remaining extra spaces and turn output into one long line of text
    (r"^\s+", r""),
    (r"\s+", r" "),
    (r" $", r""),
    # Lowercase all
    (r"^.*$", lowercase),
]

TRANSLATIONS = {
    "à": "a",
    "á": "a",
    "â": "a",
    "ã": "a",
    "é": "e",
    "è": "e",
    "ë": "e",
    "ê": "e",
    "í": "i",
    "ì": "i",
    "ï": "i",
    "î": "i",
    "ó": "o",
    "ò": "o",
    "õ": "o",
    "ô": "o",
    "ü": "u",
    "ú": "u",
    "ù": "u",
    "û": "u",
    "ý": "y",
    "ÿ": "y",
    "ç": "c",
    "ć": "c",
    "č": "c",
    "ñ": "nj",
    "ø": "ö",
    "æ": "ä",
    "š": "s",
    "ß": "ss",
    # Weird special characters encountered in Swedish text
    "ı": "i",
    "ﬁ": "fi",
    "ﬂ": "fl",
}



"""General text preprocessor that uses a regexp recipe to clean a text."""
import re
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple


class UnacceptedCharsError(Exception):
    """A custom error type for preprocessing."""

    pass


def apply(
    text: str,
    regexps: List[Tuple[str, object]],
    unaccepted_chars: str,
    translations: Optional[Dict[str, str]] = None,
) -> str:
    r"""Apply preprocessing recipe to given text.

    Result is validated by checking it for any unaccepted characters. Easiest way to define
    unaccepted characters using regexps is to use a negation of accepted chars. For example:

    ``r"[^a-zåäö \n]"``

    In addition to regexp substitutions, translation mapping is applied to input if defined.

    Args:
        text (str): input string to preprocess
        regexps (list): a preprocessing recipe in the form of regexp-substition pairs
        unaccepted_chars (str): a regexp used to search unaccepted characters
        translations (dict, optional): a translation mapping

    Raises:
        UnacceptedCharsError: raised if preprocessing fails to remove unwanted characters

    Returns:
        str: preprocessed text
    """
    result = text

    for regexp, substitution in regexps:
        result = re.sub(regexp, substitution, result)  # type: ignore  # overload not found

    if translations:
        translation = str.maketrans(translations)
        result = result.translate(translation)

    if match := re.search(unaccepted_chars, result):
        raise UnacceptedCharsError(
            f"Found unaccepted character '{match.group(0)}' in result at index {match.start()}. "
            f"Result:\n{result}"
        )
    return result



"""Word+inflection mapping for numbers in Finnish.

You can add inflections here. Note that the names of the forms must
conform regular expression "[A-Z]+". Note that a special symbols !
and " are used for ordinal versions of 1 and 2.

Currently supported inflections:

* ERI = 0567 -> nolla viisi kuusi seitsemän
* NOM = nominatiivi (viisi)
* PAR = partitiivi (viittä)
* GEN = genetiivi (viiden)
* INE = inessiivi (viidessä)
* ELA = elatiivi (viidestä)
* ILL = illatiivi (viiteen)
* ADE = adessiivi (viidellä)
* ABL = ablatiivi (viideltä)
* ALL = allatiivi (viidelle)
* ESS = essiivi (viitenä)
* TRA = translatiivi (viideksi)
* JNOM = järjestys nominatiivi (viides)
* JPAR = järjestys partitiivi (viidettä)
* JGEN = järjestys genetiivi (viidennen)
* JINE = järjestys inessiivi (viidennessä)
* JELA = järjestys elatiivi (viidennestä)
* JILL = järjestys illatiivi (viidenteen)
* JADE = järjestys adessiivi (viidennellä)
* JABL = järjestys ablatiivi (viidenneltä)
* JALL = järjestys allatiivi (viidennelle)
* JESS = järjestys essiivi (viidentenä)
* JTRA = järjestys translatiivi (viidenneksi)
"""
# Encodings for different bases of ten (ones, tens, hunderds, etc...)
UNITS = {0: "y", 1: "k", 2: "s", 3: "t", 6: "m", 9: "r", 12: "b"}

INFLECTIONS = {
    "NOM": {
        "0": "nolla",
        "1": "yksi",
        "2": "kaksi",
        "3": "kolme",
        "4": "neljä",
        "5": "viisi",
        "6": "kuusi",
        "7": "seitsemän",
        "8": "kahdeksan",
        "9": "yhdeksän",
        "o": "toista",
        "k": "kymmenen",
        "s": "sata",
        "t": "tuhat",
        "m": "miljoona",
        "r": "miljardi",
        "b": "biljoona",
    },
    "PAR": {
        "1": "yhtä",
        "2": "kahta",
        "3": "kolmea",
        "4": "neljää",
        "5": "viittä",
        "6": "kuutta",
        "7": "seitsemää",
        "8": "kahdeksaa",
        "9": "yhdeksää",
        "o": "toista",
        "k": "kymmentä",
        "s": "sataa",
        "t": "tuhatta",
        "m": "miljoonaa",
        "r": "miljardia",
        "b": "biljoonaa",
    },
    "GEN": {
        "1": "yhden",
        "2": "kahden",
        "3": "kolmen",
        "4": "neljän",
        "5": "viiden",
        "6": "kuuden",
        "7": "seitsemän",
        "8": "kahdeksan",
        "9": "yhdeksän",
        "o": "toista",
        "k": "kymmenen",
        "s": "sadan",
        "t": "tuhannen",
        "m": "miljoonan",
        "r": "miljardin",
        "b": "biljoonan",
    },
    "INE": {
        "1": "yhdessä",
        "2": "kahdessa",
        "3": "kolmessa",
        "4": "neljässä",
        "5": "viidessä",
        "6": "kuudessa",
        "7": "seitsemässä",
        "8": "kahdeksassa",
        "9": "yhdeksässä",
        "o": "toista",
        "k": "kymmenessä",
        "s": "sadassa",
        "t": "tuhannessa",
        "m": "miljoonassa",
        "r": "miljardissa",
        "b": "biljoonassa",
    },
    "ELA": {
        "1": "yhdestä",
        "2": "kahdesta",
        "3": "kolmesta",
        "4": "neljästä",
        "5": "viidestä",
        "6": "kuudesta",
        "7": "seitsemästä",
        "8": "kahdeksasta",
        "9": "yhdeksästä",
        "o": "toista",
        "k": "kymmenestä",
        "s": "sadasta",
        "t": "tuhannesta",
        "m": "miljoonasta",
        "r": "miljardista",
        "b": "biljoonasta",
    },
    "ILL": {
        "1": "yhteen",
        "2": "kahteen",
        "3": "kolmeen",
        "4": "neljään",
        "5": "viiteen",
        "6": "kuuteen",
        "7": "seitsemään",
        "8": "kahdeksaan",
        "9": "yhdeksään",
        "o": "toista",
        "k": "kymmeneen",
        "s": "sataan",
        "t": "tuhanteen",
        "m": "miljoonaan",
        "r": "miljardiin",
        "b": "biljoonaan",
    },
    "ADE": {
        "1": "yhdellä",
        "2": "kahdella",
        "3": "kolmella",
        "4": "neljällä",
        "5": "viidellä",
        "6": "kuudella",
        "7": "seitsemällä",
        "8": "kahdeksalla",
        "9": "yhdeksällä",
        "o": "toista",
        "k": "kymmenellä",
        "s": "sadalla",
        "t": "tuhannella",
        "m": "miljoonalla",
        "r": "miljardilla",
        "b": "biljoonalla",
    },
    "ABL": {
        "1": "yhdeltä",
        "2": "kahdelta",
        "3": "kolmelta",
        "4": "neljältä",
        "5": "viideltä",
        "6": "kuudelta",
        "7": "seitsemältä",
        "8": "kahdeksalta",
        "9": "yhdeksältä",
        "o": "toista",
        "k": "kymmeneltä",
        "s": "sadalta",
        "t": "tuhannelta",
        "m": "miljoonalta",
        "r": "miljardilta",
        "b": "biljoonalta",
    },
    "ALL": {
        "1": "yhdelle",
        "2": "kahdelle",
        "3": "kolmelle",
        "4": "neljälle",
        "5": "viidelle",
        "6": "kuudelle",
        "7": "seitsemälle",
        "8": "kahdeksalle",
        "9": "yhdeksälle",
        "o": "toista",
        "k": "kymmenelle",
        "s": "sadalle",
        "t": "tuhannelle",
        "m": "miljoonalle",
        "r": "miljardille",
        "b": "biljoonalle",
    },
    "ESS": {
        "1": "yhtenä",
        "2": "kahtena",
        "3": "kolmena",
        "4": "neljänä",
        "5": "viitenä",
        "6": "kuutena",
        "7": "seitsemänä",
        "8": "kahdeksana",
        "9": "yhdeksänä",
        "o": "toista",
        "k": "kymmenenä",
        "s": "satana",
        "t": "tuhantena",
        "m": "miljoonana",
        "r": "miljardina",
        "b": "biljoonana",
    },
    "TRA": {
        "1": "yhdeksi",
        "2": "kahdeksi",
        "3": "kolmeksi",
        "4": "neljäksi",
        "5": "viideksi",
        "6": "kuudeksi",
        "7": "seitsemäksi",
        "8": "kahdeksaksi",
        "9": "yhdeksäksi",
        "o": "toista",
        "k": "kymmeneksi",
        "s": "sadaksi",
        "t": "tuhanneksi",
        "m": "miljoonaksi",
        "r": "miljardiksi",
        "b": "biljoonaksi",
    },
    "JNOM": {
        "!": "ensimmäinen",
        '"': "toinen",
        "1": "yhdes",
        "2": "kahdes",
        "3": "kolmas",
        "4": "neljäs",
        "5": "viides",
        "6": "kuudes",
        "7": "seitsemäs",
        "8": "kahdeksas",
        "9": "yhdeksäs",
        "o": "toista",
        "k": "kymmenes",
        "s": "sadas",
        "t": "tuhannes",
        "m": "miljoonas",
        "r": "miljardis",
        "b": "biljoonas",
    },
    "JPAR": {
        "!": "ensimmäistä",
        '"': "toista",
        "1": "yhdettä",
        "2": "kahdetta",
        "3": "kolmatta",
        "4": "neljättä",
        "5": "viidettä",
        "6": "kuudetta",
        "7": "seitsemättä",
        "8": "kahdeksatta",
        "9": "yhdeksättä",
        "o": "toista",
        "k": "kymmenettä",
        "s": "sadatta",
        "t": "tuhannetta",
        "m": "miljoonatta",
        "r": "miljarditta",
        "b": "biljoonatta",
    },
    "JGEN": {
        "!": "ensimmäisen",
        '"': "toisen",
        "1": "yhdennen",
        "2": "kahdennen",
        "3": "kolmannen",
        "4": "neljännen",
        "5": "viidennen",
        "6": "kuudennen",
        "7": "seitsemännen",
        "8": "kahdeksannen",
        "9": "yhdeksännen",
        "o": "toista",
        "k": "kymmenennen",
        "s": "sadannen",
        "t": "tuhannennen",
        "m": "miljoonannen",
        "r": "miljardinnen",
        "b": "biljoonannen",
    },
    "JINE": {
        "!": "ensimmäisessä",
        '"': "toisessa",
        "1": "yhdennessä",
        "2": "kahdennessa",
        "3": "kolmannessa",
        "4": "neljännessä",
        "5": "viidennessä",
        "6": "kuudennessa",
        "7": "seitsemännessä",
        "8": "kahdeksannessa",
        "9": "yhdeksännessä",
        "o": "toista",
        "k": "kymmenennessä",
        "s": "sadannessa",
        "t": "tuhannennessa",
        "m": "miljoonannessa",
        "r": "miljardinnessa",
        "b": "biljoonannessa",
    },
    "JELA": {
        "!": "ensimmäisestä",
        '"': "toisesta",
        "1": "yhdennestä",
        "2": "kahdennesta",
        "3": "kolmannesta",
        "4": "neljännestä",
        "5": "viidennestä",
        "6": "kuudennesta",
        "7": "seitsemännestä",
        "8": "kahdeksannesta",
        "9": "yhdeksännestä",
        "o": "toista",
        "k": "kymmenennestä",
        "s": "sadannesta",
        "t": "tuhannennesta",
        "m": "miljoonannesta",
        "r": "miljardinnesta",
        "b": "biljoonannesta",
    },
    "JILL": {
        "!": "ensimmäiseen",
        '"': "toiseen",
        "1": "yhdenteen",
        "2": "kahdenteen",
        "3": "kolmanteen",
        "4": "neljänteen",
        "5": "viidenteen",
        "6": "kuudenteen",
        "7": "seitsemänteen",
        "8": "kahdeksanteen",
        "9": "yhdeksänteen",
        "o": "toista",
        "k": "kymmenenteen",
        "s": "sadanteen",
        "t": "tuhannenteen",
        "m": "miljoonanteen",
        "r": "miljardinteen",
        "b": "biljoonanteen",
    },
    "JADE": {
        "!": "ensimmäisellä",
        '"': "toisella",
        "1": "yhdennellä",
        "2": "kahdennella",
        "3": "kolmannella",
        "4": "neljännellä",
        "5": "viidennellä",
        "6": "kuudennella",
        "7": "seitsemännellä",
        "8": "kahdeksannella",
        "9": "yhdeksännellä",
        "o": "toista",
        "k": "kymmenennellä",
        "s": "sadannella",
        "t": "tuhannennella",
        "m": "miljoonannella",
        "r": "miljardinnella",
        "b": "biljoonannella",
    },
    "JABL": {
        "!": "ensimmäiseltä",
        '"': "toiselta",
        "1": "yhdenneltä",
        "2": "kahdennelta",
        "3": "kolmannelta",
        "4": "neljänneltä",
        "5": "viidenneltä",
        "6": "kuudennelta",
        "7": "seitsemänneltä",
        "8": "kahdeksannelta",
        "9": "yhdeksänneltä",
        "o": "toista",
        "k": "kymmenenneltä",
        "s": "sadannelta",
        "t": "tuhannennelta",
        "m": "miljoonannelta",
        "r": "miljardinnelta",
        "b": "biljoonannelta",
    },
    "JALL": {
        "!": "ensimmäiselle",
        '"': "toiselle",
        "1": "yhdennelle",
        "2": "kahdennelle",
        "3": "kolmannelle",
        "4": "neljännelle",
        "5": "viidennelle",
        "6": "kuudennelle",
        "7": "seitsemännelle",
        "8": "kahdeksannelle",
        "9": "yhdeksännelle",
        "o": "toista",
        "k": "kymmenennelle",
        "s": "sadannelle",
        "t": "tuhannennelle",
        "m": "miljoonannelle",
        "r": "miljardinnelle",
        "b": "biljoonannelle",
    },
    "JESS": {
        "!": "ensimmäisenä",
        '"': "toisena",
        "1": "yhdentenä",
        "2": "kahdentena",
        "3": "kolmantena",
        "4": "neljäntenä",
        "5": "viidentenä",
        "6": "kuudentena",
        "7": "seitsemäntenä",
        "8": "kahdeksantena",
        "9": "yhdeksäntenä",
        "o": "toista",
        "k": "kymmenentenä",
        "s": "sadantena",
        "t": "tuhannentena",
        "m": "miljoonantena",
        "r": "miljardintena",
        "b": "biljoonantena",
    },
    "JTRA": {
        "!": "ensimmäiseksi",
        '"': "toiseksi",
        "1": "yhdenneksi",
        "2": "kahdenneksi",
        "3": "kolmanneksi",
        "4": "neljänneksi",
        "5": "viidenneksi",
        "6": "kuudenneksi",
        "7": "seitsemänneksi",
        "8": "kahdeksanneksi",
        "9": "yhdeksänneksi",
        "o": "toista",
        "k": "kymmenenneksi",
        "s": "sadanneksi",
        "t": "tuhannenneksi",
        "m": "miljoonanneksi",
        "r": "miljardinneksi",
        "b": "biljoonanneksi",
    },
}



"""Number expansion for Finnish.

This package is used to convert numbers to written Finnish
form with the desired inflectional form.

Inflection mapping is defined in the module init.
"""
import re
from typing import Match

from aalto_asr_preprocessor.fi.numbers import INFLECTIONS
from aalto_asr_preprocessor.fi.numbers import UNITS


def to_nominal(match: Match[str]) -> str:
    """Regexp replacing function for expanding a string of single digits.

    Args:
        match (Match): captured digit

    Returns:
        str: word corresponding to the digit
    """
    digit = match.group(0)
    return INFLECTIONS["NOM"][digit] + " "


def inflect(match: Match[str]) -> str:
    """Regexp replacing function for inflecting a digit in correct inflectional form.

    Args:
        match (Match): captured digit and inflectional form (e.g. ('8','TRA'))

    Returns:
        str: word corresponding to the digit in correct case (e.g. 'kahdeksaksi')
    """
    digit = match.group(1)
    form = match.group(2)
    return INFLECTIONS[form][digit]


def encode_base_ten(digits: str) -> str:
    """Encode bases of ten in a number string for expansion.

    Args:
        digits (str): a number to encode (e.g. 87410)

    Returns:
        str: number encoded with bases of ten (e.g. 8k7yt 4s1k0y)
    """
    encoding = ""
    for i, d in zip(reversed(range(len(digits))), digits):
        encoding += f"{d}{UNITS[i%3]}"
        if i > 2 and i % 3 == 0:
            encoding += f"{UNITS[i]} "
    return encoding


def expand(match: Match[str]) -> str:
    """Expand a number string in Finnish text to equivalent word representation.

    Args:
        match (Match): captured number and optionally its inflectional form: 100#PAR

    Returns:
        str: the matched number(s) expanded as words
    """
    digits = match.group(1)
    form = match.group(2)
    if not form:
        form = "NOM"

    if form == "ERI" or digits.startswith("0") or len(digits) > 13:
        return str(re.sub(r"[0-9]", to_nominal, digits).strip())

    result = encode_base_ten(digits)

    # Convert teens
    result = re.sub(r"1k(\d)y", r"\1o", result)
    result = re.sub(r"0o", r"1k", result)
    # Remove redundant zeros and ones
    result = re.sub(r"0[a-z]\s*", r"", result)
    result = re.sub(r"y", r"", result)
    result = re.sub(r"\s[a-z]", r"", result)
    result = re.sub(r"1([kstmrb])", r"\1", result)
    # Add inflectional forms
    result = re.sub(r"(\S)", r"\1" + form, result)
    result = re.sub(r"(?<=NOM|PAR)([kstmrb])NOM", r"\1PAR", result)
    # Fix ordinals "ensimmäinen" and "toinen"
    result = re.sub(r"1(J[A-Z]+)$", r"!\1", result)
    result = re.sub(r"2(J[A-Z]+)$", r'"\1', result)

    # Inflect encoded string
    result = re.sub(r'([0-9a-z!"]{1})([A-Z]{3,4})', inflect, result)

    return result.replace(" ", "")



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['aalto_asr_preprocessor',
 'aalto_asr_preprocessor.fi',
 'aalto_asr_preprocessor.fi.numbers']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0']

entry_points = \
{'console_scripts': ['aalto-prep = aalto_asr_preprocessor.__main__:main']}

setup_kwargs = {
    'name': 'aalto-asr-preprocessor',
    'version': '2021.3.10',
    'description': 'Aalto ASR preprocessing tool for preparing texts.',
    'long_description': "Aalto ASR preprocessing package\n===============================\n\n|License| |Tests|\n\n|pre-commit| |Black|\n\n.. |License| image:: https://img.shields.io/github/license/aalto-speech/aalto-asr-preprocessor\n   :target: https://opensource.org/licenses/MIT\n   :alt: License\n.. |Tests| image:: https://github.com/aalto-speech/aalto-asr-preprocessor/workflows/Tests/badge.svg\n   :target: https://github.com/aalto-speech/aalto-asr-preprocessor/actions?workflow=Tests\n   :alt: Tests\n.. |pre-commit| image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n.. |Black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Black\n\nFeatures\n--------\n\n* Rule-based preprocessing recipes for ASR\n* Recipes are validated with tests\n\nRequirements\n------------\n\n* Python >= 3.8\n* Click\n\nInstallation\n------------\n\nTo install the Aalto ASR preprocessor,\nclone the repository and run this command in your terminal:\n\n.. code-block:: console\n\n   $ git clone https://github.com/aalto-speech/aalto-asr-preprocessor.git\n   $ cd aalto-asr-preprocessor\n   $ pip install .\n\nUsage\n-----\n\nFor detailed instructions, see `Usage`_\nor type ``aalto-prep --help`` in terminal.\n\nContributing\n------------\n\nContributions are very welcome.\nTo learn more, see the `Contributor Guide`_.\n\nLicense\n-------\n\nDistributed under the terms of the MIT_ license,\n*Aalto ASR preprocessor* is free and open source software.\n\nIssues\n------\n\nIf you encounter any problems,\nplease `file an issue`_ along with a detailed description.\n\nCredits\n-------\n\nThis project uses `@cjolowicz`_'s `Hypermodern Python Cookiecutter`_ template.\n\n\n.. _@cjolowicz: https://github.com/cjolowicz\n.. _MIT: http://opensource.org/licenses/MIT\n.. _Hypermodern Python Cookiecutter: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n.. _file an issue: https://github.com/aalto-speech/aalto-asr-preprocessor/issues\n.. _Contributor Guide: CONTRIBUTING.rst\n.. _Usage: docs/index.rst\n",
    'author': 'Anja Virkkunen',
    'author_email': 'anja.virkkunen@aalto.fi',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/aalto-speech/aalto-asr-preprocessor',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Simple command line client for the package.

The client expects three arguments: input, output, and a recipe file.
Run ``aalto-prep --help`` to see more.
"""
from importlib.abc import Loader
from importlib.util import module_from_spec
from importlib.util import spec_from_file_location
from typing import Any

import click

import aalto_asr_preprocessor.preprocessor as prep


@click.command()
@click.argument("input", type=click.File())
@click.argument("output", type=click.File(mode="w"))
@click.argument("recipefile", type=click.Path(exists=True))
@click.option(
    "--add-linefeed",
    is_flag=True,
    help="Add line feeds to keep your input lines separated.",
)
@click.version_option()
def main(input: Any, output: Any, recipefile: Any, add_linefeed: Any) -> None:
    """Preprocess INPUT to OUTPUT using RECIPEFILE."""
    line_separator = ""
    if add_linefeed:
        line_separator = "\n"

    spec = spec_from_file_location(recipefile, recipefile)
    assert isinstance(spec.loader, Loader)  # for mypy
    recipe = module_from_spec(spec)
    spec.loader.exec_module(recipe)

    texts = input.readlines()
    for text in texts:
        try:
            result = prep.apply(
                text, recipe.REGEXPS, recipe.UNACCEPTED_CHARS, recipe.TRANSLATIONS  # type: ignore
            )
        except AttributeError:
            result = prep.apply(text, recipe.REGEXPS, recipe.UNACCEPTED_CHARS)  # type: ignore
        output.write(result + line_separator)


if __name__ == "__main__":
    main(prog_name="aalto_asr_preprocessor")  # pragma: no cover
